# define MAXDIM  40
# define SZ_BUF  79

typedef enum 
{
  tAny,

  tBool,
  tInt8, tUInt8,
  tInt16, tUInt16,
  tInt32, 
  tInt64,
  tFloat32, tFloat64,
  tComplex32, tComplex64,

  tDefault = tFloat64,
#if HAS_INT64
  tLong = tInt64
#else
  tLong = tInt32
#endif
} NumericType;
#define nNumericType (tComplex64+1)

typedef struct {
	void *data;			/* pointer to data buffer */
	void *imag;			/* imaginary part, or NULL */
	long ndim;			/* dimension (rank) of array */
	long nelem;			/* total number of elements */
	long shape[MAXDIM];		/* length of each axis */
	long stride[MAXDIM];
	long byteoffset;
	long bytestride;
	long itemsize;
	char type[SZ_BUF+1];
	/* flags */
	int C_array;		/* aligned, contiguous, and not byteswapped */
	int writeable;
	int aligned;
	int contiguous;
	int byteswap;
        int ntype;              /* NumericType */
	Float64 temp;
        char *wptr;
	int byteorder;           
} NDInfo;

typedef enum 
{
	NUM_LITTLE_ENDIAN=0,
	NUM_BIG_ENDIAN = 1
} NDInfoByteOrder;

typedef enum
{
	NUM_CONTIGUOUS=1,
	NUM_NOTSWAPPED=2,
	NUM_ALIGNED=4
} NumRequirements;

#define C_ARRAY (NUM_CONTIGUOUS | NUM_NOTSWAPPED | NUM_ALIGNED)

#define NA_Begin() Py_Initialize(); import_libnumarray();
#define NA_End()   NA_Done(); Py_Finalize();

#define NA_OFFSETDATA(info) ((void *) (((char*)(info).data)+(info).byteoffset))

/* unaligned NA_COPY functions */
#define NA_COPY1(i, o)  (*(o) = *(i))
#define NA_COPY2(i, o)  NA_COPY1(i, o), NA_COPY1(i+1, o+1)
#define NA_COPY4(i, o)  NA_COPY2(i, o), NA_COPY2(i+2, o+2)
#define NA_COPY8(i, o)  NA_COPY4(i, o), NA_COPY4(i+4, o+4)
#define NA_COPY16(i, o) NA_COPY8(i, o), NA_COPY8(i+8, o+8)

/* byteswapping macros: these fail if i==o */
#define NA_SWAP1(i, o)  NA_COPY1(i, o)
#define NA_SWAP2(i, o)  NA_SWAP1(i, o+1), NA_SWAP1(i+1, o)
#define NA_SWAP4(i, o)  NA_SWAP2(i, o+2), NA_SWAP2(i+2, o)
#define NA_SWAP8(i, o)  NA_SWAP4(i, o+4), NA_SWAP4(i+4, o)
#define NA_SWAP16(i, o) NA_SWAP8(i, o+8), NA_SWAP8(i+8, o)

/* complex byteswaps must swap each part (real, imag) independently */
#define NA_COMPLEX_SWAP8(i, o) NA_SWAP4(i, o), NA_SWAP4(i+4, o+4)
#define NA_COMPLEX_SWAP16(i, o) NA_SWAP8(i, o), NA_SWAP8(i+8, o+8)

/* byteswapping macros:  these work even if i == o */
#define NA_TSWAP1(i, o, t) NA_COPY1(i, t), NA_SWAP1(t, o)
#define NA_TSWAP2(i, o, t) NA_COPY2(i, t), NA_SWAP2(t, o)
#define NA_TSWAP4(i, o, t) NA_COPY4(i, t), NA_SWAP4(t, o)
#define NA_TSWAP8(i, o, t) NA_COPY8(i, t), NA_SWAP8(t, o)

/* fast copy functions for %N aligned i and o */
#define NA_ACOPY1(i, o) (((Int8    *)o)[0]   = ((Int8    *)i)[0])
#define NA_ACOPY2(i, o) (((Int16   *)o)[0]   = ((Int16   *)i)[0])
#define NA_ACOPY4(i, o) (((Int32   *)o)[0]   = ((Int32   *)i)[0])
#define NA_ACOPY8(i, o) (((Float64 *)o)[0]   = ((Float64 *)i)[0])
#define NA_ACOPY16(i, o) (((Complex64 *)o)[0]   = ((Complex64 *)i)[0])

/* from here down, type("ai") is NDInfo*  */

#define NA_PTR(ai)   ((char *) NA_OFFSETDATA(*(ai)))
#define NA_TEMP(ai)  ((char *) &(ai)->temp)
#define NA_PTR1(ai, i)       (NA_PTR(ai) +                                   \
                              (i)*(ai)->stride[0])
#define NA_PTR2(ai, i, j)    (NA_PTR(ai) +                                   \
                              (i)*(ai)->stride[0] +                          \
                              (j)*(ai)->stride[1])
#define NA_PTR3(ai, i, j, k) (NA_PTR(ai) +                                   \
                              (i)*(ai)->stride[0] +                          \
                              (j)*(ai)->stride[1] +                          \
                              (k)*(ai)->stride[2])
#define NA_RESULT(ai, type)   (*((type *) NA_TEMP(ai)))
#define NA_SET_TEMP(ai, type, v) (NA_RESULT(ai, type) = ((type) (v)))

#define NA_SWAPFloat64   NA_SWAP8
#define NA_SWAPFloat32   NA_SWAP4
#define NA_SWAPInt64     NA_SWAP8
#define NA_SWAPInt32     NA_SWAP4
#define NA_SWAPInt16     NA_SWAP2
#define NA_SWAPUInt16    NA_SWAP2
#define NA_SWAPInt8      NA_SWAP1
#define NA_SWAPUInt8     NA_SWAP1
#define NA_SWAPBool      NA_SWAP1

#define NA_COPYFloat64   NA_COPY8
#define NA_COPYFloat32   NA_COPY4
#define NA_COPYInt64     NA_COPY8
#define NA_COPYInt32     NA_COPY4
#define NA_COPYInt16     NA_COPY2
#define NA_COPYUInt16    NA_COPY2
#define NA_COPYInt8      NA_COPY1
#define NA_COPYUInt8     NA_COPY1
#define NA_COPYBool      NA_COPY1

#define NA_REM(ai, ptr)  ((ai)->wptr = (ptr))
#define NA_REC(ai)       ((ai)->wptr)

/* ========================== ptr get/set ================================ */

/* byteswapping */
#define NA_GETPb(ai, type, ptr)                                               \
         (NA_REM(ai, ptr), NA_SWAP##type(NA_REC(ai), NA_TEMP(ai)),            \
	  NA_RESULT(ai, type))        

/* aligning */
#define NA_GETPa(ai, type, ptr)                                               \
         (NA_REM(ai, ptr), NA_COPY##type(NA_REC(ai), NA_TEMP(ai)),            \
	  NA_RESULT(ai, type))

/* fast (aligned, !byteswapped) */
#define NA_GETPf(ai, type, ptr) (*((type *) ptr))

#define NA_GETP(ai, type, ptr)                                                \
   ((ai)->byteswap ? NA_GETPb(ai, type, ptr)                                  \
                   : ((ai)->aligned ? NA_GETPf(ai, type, ptr)                 \
                                    : NA_GETPa(ai, type, ptr)))

/* NOTE:  NA_SET* macros cannot be used as values. */

/* byteswapping */
#define NA_SETPb(ai, type, ptr, v)                                            \
         (NA_REM(ai, ptr), NA_SET_TEMP(ai, type, v),                          \
	  NA_SWAP##type(NA_TEMP(ai), NA_REC(ai)))

/* aligning */
#define NA_SETPa(ai, type, ptr, v)                                            \
         (NA_REM(ai, ptr), NA_SET_TEMP(ai, type, v),                          \
          NA_COPY##type(NA_TEMP(ai), NA_REC(ai)))

/* fast (aligned, !byteswapped) */
#define NA_SETPf(ai, type, ptr, v) ((*((type *) ptr)) = (v))

#define NA_SETP(ai, type, ptr, v)                                             \
   ((ai)->byteswap ? NA_SETPb((ai), type, (ptr), (v))                         \
                   : ((ai)->aligned ? NA_SETPf((ai), type, (ptr), (v))        \
                                    : NA_SETPa((ai), type, (ptr), (v))))

/* ========================== 1 index get/set ============================ */

/* byteswapping */
#define NA_GET1b(ai, type, i)    NA_GETPb(ai, type, NA_PTR1(ai, i))
/* aligning */
#define NA_GET1a(ai, type, i)    NA_GETPa(ai, type, NA_PTR1(ai, i))
/* fast (aligned, !byteswapped) */
#define NA_GET1f(ai, type, i)    NA_GETPf(ai, type, NA_PTR1(ai, i))
/* testing */
#define NA_GET1(ai, type, i)     NA_GETP(ai, type, NA_PTR1(ai, i))

/* byteswapping */
#define NA_SET1b(ai, type, i, v) NA_SETPb(ai, type, NA_PTR1(ai, i), v)
/* aligning */
#define NA_SET1a(ai, type, i, v) NA_SETPa(ai, type, NA_PTR1(ai, i), v)
/* fast (aligned, !byteswapped) */
#define NA_SET1f(ai, type, i, v) NA_SETPf(ai, type, NA_PTR1(ai, i), v)
/* testing */
#define NA_SET1(ai, type, i, v)  NA_SETP(ai, type,  NA_PTR1(ai, i), v)

/* ========================== 2 index get/set ============================= */

/* byteswapping */
#define NA_GET2b(ai, type, i, j)    NA_GETPb(ai, type, NA_PTR2(ai, i, j))
/* aligning */
#define NA_GET2a(ai, type, i, j)    NA_GETPa(ai, type, NA_PTR2(ai, i, j))
/* fast (aligned, !byteswapped) */
#define NA_GET2f(ai, type, i, j)    NA_GETPf(ai, type, NA_PTR2(ai, i, j))
/* testing */
#define NA_GET2(ai, type, i, j)     NA_GETP(ai, type, NA_PTR2(ai, i, j))

/* byteswapping */
#define NA_SET2b(ai, type, i, j, v) NA_SETPb(ai, type, NA_PTR2(ai, i, j), v)
/* aligning */
#define NA_SET2a(ai, type, i, j, v) NA_SETPa(ai, type, NA_PTR2(ai, i, j), v)
/* fast (aligned, !byteswapped) */
#define NA_SET2f(ai, type, i, j, v) NA_SETPf(ai, type, NA_PTR2(ai, i, j), v)

#define NA_SET2(ai, type, i, j,  v)  NA_SETP(ai, type,  NA_PTR2(ai, i, j), v)

/* ========================== 3 index get/set ============================= */

/* byteswapping */
#define NA_GET3b(ai, type, i, j, k)    NA_GETPb(ai, type, NA_PTR3(ai, i, j, k))
/* aligning */
#define NA_GET3a(ai, type, i, j, k)    NA_GETPa(ai, type, NA_PTR3(ai, i, j, k))
/* fast (aligned, !byteswapped) */
#define NA_GET3f(ai, type, i, j, k)    NA_GETPf(ai, type, NA_PTR3(ai, i, j, k))
/* testing */
#define NA_GET3(ai, type, i, j, k)     NA_GETP(ai, type, NA_PTR3(ai, i, j, k))

/* byteswapping */
#define NA_SET3b(ai, type, i, j, k, v)                                        \
        NA_SETPb(ai, type, NA_PTR3(ai, i, j, k), v)
/* aligning */
#define NA_SET3a(ai, type, i, j, k, v)                                        \
        NA_SETPa(ai, type, NA_PTR3(ai, i, j, k), v)
/* fast (aligned, !byteswapped) */
#define NA_SET3f(ai, type, i, j, k, v)                                        \
        NA_SETPf(ai, type, NA_PTR3(ai, i, j, k), v)
#define NA_SET3(ai, type, i, j, k, v)                                         \
        NA_SETP(ai, type,  NA_PTR3(ai, i, j, k), v)

/* ========================== 1D get/set ================================== */

#define NA_GET1Db(ai, type, base, cnt, out)                                   \
        { int i, stride = ai->stride[ai->ndim-1];                             \
           for(i=0; i<cnt; i++) {                                             \
               out[i] = NA_GETPb(ai, type, base);                             \
               base += stride;                                                \
           }                                                                  \
        } 

#define NA_GET1Da(ai, type, base, cnt, out)                                   \
        { int i, stride = ai->stride[ai->ndim-1];                             \
           for(i=0; i<cnt; i++) {                                             \
               out[i] = NA_GETPa(ai, type, base);                             \
               base += stride;                                                \
           }                                                                  \
        } 

#define NA_GET1Df(ai, type, base, cnt, out)                                   \
        { int i, stride = ai->stride[ai->ndim-1];                             \
           for(i=0; i<cnt; i++) {                                             \
               out[i] = NA_GETPf(ai, type, base);                             \
               base += stride;                                                \
           }                                                                  \
        } 

#define NA_GET1D(ai, type, base, cnt, out)                                    \
        if (ai->byteswap) {                                                   \
              NA_GET1Db(ai, type, base, cnt, out);                            \
        } else if (!ai->aligned) {                                            \
              NA_GET1Da(ai, type, base, cnt, out);                            \
        } else {                                                              \
	      NA_GET1Df(ai, type, base, cnt, out);                            \
	}
        
#define NA_SET1Db(ai, type, base, cnt, in)                                    \
        { int i, stride = ai->stride[ai->ndim-1];                             \
           for(i=0; i<cnt; i++) {                                             \
               NA_SETPb(ai, type, base, in[i]);                               \
               base += stride;                                                \
           }                                                                  \
        }

#define NA_SET1Da(ai, type, base, cnt, in)                                    \
        { int i, stride = ai->stride[ai->ndim-1];                             \
           for(i=0; i<cnt; i++) {                                             \
               NA_SETPa(ai, type, base, in[i]);                               \
               base += stride;                                                \
           }                                                                  \
        }

#define NA_SET1Df(ai, type, base, cnt, in)                                    \
        { int i, stride = ai->stride[ai->ndim-1];                             \
           for(i=0; i<cnt; i++) {                                             \
               NA_SETPf(ai, type, base, in[i]);                               \
               base += stride;                                                \
           }                                                                  \
        }

#define NA_SET1D(ai, type, base, cnt, out)                                    \
        if (ai->byteswap) {                                                   \
              NA_SET1Db(ai, type, base, cnt, out);                            \
        } else if (!ai->aligned) {                                            \
              NA_SET1Da(ai, type, base, cnt, out);                            \
        } else {                                                              \
	      NA_SET1Df(ai, type, base, cnt, out);                            \
	}
        
/* ========================== utilities ================================== */

#define MIN(x,y) (((x)<=(y)) ? (x) : (y))
#define MAX(x,y) (((x)>=(y)) ? (x) : (y))
#define ELEM(x)  (sizeof(x)/sizeof(x[0]))
